import { Inject, Injectable } from '@nestjs/common';
import { ClientProxy } from '@nestjs/microservices';
import { interval, Observable, Subject } from 'rxjs';
import * as net from 'net'
import { ConfigService } from '@nestjs/config';
import { connectToAfisJava } from 'libs/java/afis.utils';
import { Fingerprint, FingerprintPayload, FingerprintPayloadUI, FisMessage, JavaResponse, PersonFingerprintData, TransparencyResponseData } from 'libs';
import { v1 as uuid } from 'uuid'

@Injectable()
export class FisVerificationService {
  private port: number
  private host: string
  private edgeScoreThreshold: number
  private javaClient: net.Socket
  private incomingMessageFromJava!: Subject<JavaResponse>
  private registeredFingerprintData: Fingerprint[] = []

  constructor(
    @Inject(`SAMPLEAPP_SERVICE`) private sampleClient: ClientProxy,
    @Inject(`FINGERPRINT_SERVICE`) private fingerprintClient: ClientProxy,
    private configService: ConfigService
  ) {
    // Set up incoming responses or messages from TCP Java
    this.incomingMessageFromJava = new Subject()
    this.port = this.configService.get<number>('afis.tcpPort') as number
    this.host = this.configService.get<string>('afis.host') as string
    this.edgeScoreThreshold = this.configService.get<number>('edgeScore.mark') as number

    this.handleJavaResponse()

    this.javaClient = connectToAfisJava(this.host, this.port)
    this.javaClient.on(`data`, (data) => {
      let message: JavaResponse = JSON.parse(data.toString())
      this.incomingMessageFromJava.next(message)
    })
  }

  public handleMessage(message: FisMessage): void {
    console.log(`Sending fingerprint data to Java verification....`)
    let payload: FingerprintPayloadUI = message.data as FingerprintPayloadUI
    let command: FingerprintPayload = {
      ...payload,
      fpTemplateArray: this.registeredFingerprintData
    }
    this.javaClient.write(JSON.stringify(command) + `\n`)
  }

  private populateRegisteredFpData(): void {
    // let command: FingerprintPayload = {
    //   id: uuid(),
    //   cmd: `Populate Registered Prints`,
    //   date: new Date(),
    //   personInfo: null,
    //   fpScan: ``,
    //   fingerPosition: 0,
    //   fpTemplateArray: this.registeredFingerprintData
    // }
    // this.javaClient.write(JSON.stringify(command))
  }

  private handleJavaResponse(): void {
    this.incomingMessageFromJava.subscribe(message => {
      // console.log(`Java server response: ${message.message}`)
      if (message.operation == `Quality Assurance`) {
        let response: FisMessage = {
          header: {
            messageID: uuid(),
            messageName: `Verification Response`
          },
          data: {
            id: uuid(),
            cmd: `Quality Assurance`,
            message: `${message.data.name} finger position ${message.data.fingerprints[0].fpPosition} edge score is ${message.edgeScore > this.edgeScoreThreshold ? `good` : `bad`}`,
            data: {
              score: message.score,
              person: message.data.name,
              fpPosition: message.data.fingerprints[0].fpPosition,
              fpQuality: message.edgeScore > this.edgeScoreThreshold ? `good` : `bad`
            } as TransparencyResponseData
          }
        }
        // console.log(`Sending back to fingerpint nest`)
        console.log(response)
        this.fingerprintClient.emit(`verification`, response)
      }

      if (message.operation == `Registration` && message.status == `Success`) {
        this.registeredFingerprintData.push((message.data as PersonFingerprintData).fingerprints[0] as Fingerprint)
        console.log(message.data)
        // console.log(this.registeredFingerprintData)
        console.log(`${this.registeredFingerprintData.length} registered print.`)
      }
    })
  }

}

/* Reason based on ChatGPT: Why ClientProxy.send() or emit() won't work 
The NestJS ClientProxy (when configured for Transport.TCP) uses its own custom binary protocol under the hood. This includes:
A message envelope that wraps your pattern and data.
Message IDs and event headers.
A specific serialization/deserialization logic that is not plain JSON.
NO newline character at the end.
So when your Java server, which expects a newline-terminated JSON message, receives this custom binary format, it gets confused or sees null.*/

/* data: {
messageData: string
messageTemplateData: string
deviceInfo: object} */